@@ -1,6 +1,6 @@
 /* CLI utilities.
 
-   Copyright (C) 2011-2019 Free Software Foundation, Inc.
+   Copyright (C) 2011-2022 Free Software Foundation, Inc.
 
    This file is part of GDB.
 
@@ -56,7 +56,7 @@ extern void report_unrecognized_option_error (const char *command,
 
 /* Builds the help string for a command documented by PREFIX,
    followed by the extract_info_print_args help for ENTITY_KIND.  If
-   DOCUMENT_N_FLAG is true then help text descibing the -n flag is also
+   DOCUMENT_N_FLAG is true then help text describing the -n flag is also
    included.  */
 
 const char *info_print_args_help (const char *prefix,
@@ -215,9 +215,9 @@ check_for_argument (char **str, const char *arg)
 
 struct qcs_flags
 {
-  int quiet = false;
-  int cont = false;
-  int silent = false;
+  bool quiet = false;
+  bool cont = false;
+  bool silent = false;
 };
 
 /* Validate FLAGS.  Throws an error if both FLAGS->CONT and
